package jrummikub.control;

import java.util.ArrayList;
import java.util.List;

import jrummikub.model.GameSettings;
import jrummikub.model.GameState;
import jrummikub.model.IRoundState;
import jrummikub.model.RoundState;
import jrummikub.model.Score;
import jrummikub.util.Connection;
import jrummikub.util.Event;
import jrummikub.util.IEvent;
import jrummikub.util.IListener;
import jrummikub.util.IListener1;
import jrummikub.view.IView;
import jrummikub.view.IView.BottomPanelType;

/**
 * Controls a Game, at some point including all Rounds, starts new Rounds
 */
public class GameControl {
	private SaveControl saveControl;

	private GameSettings gameSettings;
	private IView view;
	RoundControl roundControl;
	private GameState gameState;
	private List<Connection> connections = new ArrayList<Connection>();
	private Event endOfGameEvent = new Event();

	/**
	 * Constructor
	 * 
	 * @param gameSettings
	 *            the game settings
	 * @param saveControl
	 *            the save control
	 * @param view
	 *            the view
	 */
	public GameControl(GameSettings gameSettings, SaveControl saveControl,
			IView view) {
		this.gameSettings = gameSettings;
		this.saveControl = saveControl;
		this.view = view;

		gameState = new GameState();
		saveControl.setGameState(gameState);

		gameState.setFirstRoundFirstPlayer((int) (Math.random() * gameSettings
				.getPlayerList().size()));

		connections.add(view.getNewRoundEvent().add(new IListener() {
			@Override
			public void handle() {
				startRound();
			}
		}));

		connections.add(view.getNewGameEvent().add(new IListener() {
			@Override
			public void handle() {
				endGame();
			}
		}));

		connections.add(view.getEndProgramEvent().add(new IListener() {
			@Override
			public void handle() {
				endProgram();
			}
		}));
	}

	/**
	 * Is emitted when the user ends the game and start a new one
	 * 
	 * @return the endOfGameEvent
	 */
	public IEvent getEndOfGameEvent() {
		return endOfGameEvent;
	}

	private void endGame() {
		removeListeners();
		endOfGameEvent.emit();
	}

	private void removeListeners() {
		for (Connection c : connections) {
			c.remove();
		}
	}

	/**
	 * Game gets started by initializing the first Round
	 */
	public void startGame() {
		startRound();
	}

	/**
	 * Continues game after loading
	 * 
	 * @param gameState
	 *            the saved GameState (Players, startplayer, points)
	 * @param roundState
	 *            the saved RoundState (activePlayer, Table, heap etc)
	 */
	public void continueGame(GameState gameState, IRoundState roundState) {
		this.gameState = gameState;
		if (roundState == null) {
			showScorePanel();
		} else {
			prepareRound(roundState);
			roundControl.continueRound();
		}
	}

	/**
	 * Clean abortion of a running game when another one is loaded or when a new
	 * one is started
	 * 
	 */
	public void abortGame() {
		removeListeners();
		if (roundControl != null) {
			roundControl.abortRound();
			roundControl = null;
		}

		view.clearView();
	}

	private void startRound() {
		if (roundControl != null) {
			return;
		}

		view.showScorePanel(false);

		IRoundState roundState = new RoundState(gameSettings);
		prepareRound(roundState);
		roundControl.startRound();
	}

	private void prepareRound(IRoundState roundState) {
		saveControl.setRoundState(roundState);

		roundState.setActivePlayerNumber(gameState.getFirstRoundFirstPlayer()
				+ gameState.getScores().size());

		roundControl = new RoundControl(roundState, view);
		roundControl.getEndOfRoundEvent().add(new IListener1<Score>() {

			@Override
			public void handle(Score roundScore) {
				endOfRound(roundScore);
			}
		});

		roundControl.getRestartRoundEvent().add(new IListener() {

			@Override
			public void handle() {
				restartRound();
			}
		});
	}

	private void restartRound() {
		roundControl = null;
		startRound();
	}

	private void endOfRound(Score roundScore) {
		gameState.getScores().add(roundScore);
		saveControl.setRoundState(null);

		roundControl = null;
		showScorePanel();
	}

	private void showScorePanel() {
		view.setBottomPanel(BottomPanelType.WIN_PANEL);

		view.getScorePanel().setPlayers(gameSettings.getPlayerList());
		view.getScorePanel().setScores(gameState.getScores());
		view.getScorePanel().setAccumulatedScore(
				gameState.getAccumulatedScore());
		view.getScorePanel().update();
		view.showScorePanel(true);
	}

	private void endProgram() {
		System.exit(0);
	}

}
